cdef class CopyReader:
    """Read from iterable Copy object."""

    def __init__(
        self,
        object copyobj,
    ):
        """Class initialization."""

        self.copyobj = copyobj
        self.iterator = iter(self.copyobj.__enter__())
        self.bufferobj = bytearray()
        self.closed = False
        self.total_read = 0

    cpdef bytes read(self, long long size):
        """Read from copy."""

        if self.closed:
            raise RuntimeError("Copy object already closed.")

        if size <= 0:
            return b""

        cdef object chunk
        cdef Py_ssize_t buffer_len
        cdef bytes result

        try:
            while len(self.bufferobj) < size:
                chunk = next(self.iterator)
                self.bufferobj.extend(chunk)

            result = bytes(self.bufferobj[:size])
            del self.bufferobj[:size]
            self.total_read += len(result)
            return result

        except StopIteration:
            self.close()
            buffer_len = len(self.bufferobj)

            if buffer_len > 0:
                if size >= buffer_len:
                    result = bytes(self.bufferobj)
                    self.bufferobj = bytearray()
                else:
                    result = bytes(self.bufferobj[:size])
                    del self.bufferobj[:size]
                self.total_read += len(result)
                return result
            return b""

    cpdef long long tell(self):
        """Return the current stream position."""

        if self.closed:
            raise RuntimeError("Copy object already closed.")

        return self.total_read

    cpdef void close(self):
        """Close CopyReader."""

        if not self.closed:
            self.copyobj.__exit__(None, None, None)
            self.closed = True



cdef class CopyReader:

    cdef object copyobj
    cdef object iterator
    cdef object bufferobj
    cdef bint closed
    cdef long long total_read

    cpdef bytes read(self, long long size)
    cpdef long long tell(self)
    cpdef void close(self)



from .connector import PGConnector
from .copy import CopyBuffer
from .errors import (
    CopyBufferError,
    CopyBufferObjectError,
    CopyBufferTableNotDefined,
    PGPackDumperError,
    PGPackDumperReadError,
    PGPackDumperWriteError,
    PGPackDumperWriteBetweenError,
)
from .logger import DumperLogger
from .metadata import read_metadata
from .query import (
    chunk_query,
    query_path,
    query_template,
    random_name,
    search_object,
)
from .reader import CopyReader
from .stream import StreamReader
from .structs import PGObject


__all__ = (
    "CopyBuffer",
    "CopyBufferError",
    "CopyBufferObjectError",
    "CopyBufferTableNotDefined",
    "CopyReader",
    "DumperLogger",
    "PGConnector",
    "PGObject",
    "PGPackDumperError",
    "PGPackDumperReadError",
    "PGPackDumperWriteBetweenError",
    "PGPackDumperWriteError",
    "StreamReader",
    "chunk_query",
    "query_path",
    "query_template",
    "random_name",
    "read_metadata",
    "search_object",
)



from psycopg import Cursor

from .query import (
    query_template,
    random_name,
)


def read_metadata(
    cursor: Cursor,
    query: str | None = None,
    table_name: str | None = None,
) -> bytes:
    """Read metadata for query or table."""

    if not query and not table_name:
        raise ValueError()

    if query:
        session_name = random_name()
        prepare_name = f"{session_name}_prepare"
        table_name = f"{session_name}_temp"
        cursor.execute(query_template("prepare").format(
            prepare_name=prepare_name,
            query=query,
            table_name=table_name,
        ))

    cursor.execute(query_template("attributes").format(
        table_name=table_name,
    ))

    metadata: bytes = cursor.fetchone()[0]

    if query:
        cursor.execute(f"drop table if exists {table_name};")

    return metadata



"""Library for read and write PGPack format between PostgreSQL and file."""

from pgcopylib import (
    PGCopyReader,
    PGCopyWriter,
)
from pgpack import (
    CompressionMethod,
    PGPackReader,
    PGPackWriter,
)

from .common import (
    PGConnector,
    CopyBuffer,
    CopyBufferError,
    CopyBufferObjectError,
    CopyBufferTableNotDefined,
    PGPackDumperError,
    PGPackDumperReadError,
    PGPackDumperWriteError,
    PGPackDumperWriteBetweenError,
)
from .dumper import PGPackDumper
from .version import __version__


__all__ = (
    "__version__",
    "CompressionMethod",
    "CopyBuffer",
    "CopyBufferError",
    "CopyBufferObjectError",
    "CopyBufferTableNotDefined",
    "PGConnector",
    "PGCopyReader",
    "PGCopyWriter",
    "PGPackDumper",
    "PGPackDumperError",
    "PGPackDumperReadError",
    "PGPackDumperWriteError",
    "PGPackDumperWriteBetweenError",
    "PGPackReader",
    "PGPackWriter",
)
__author__ = "0xMihalich"



from enum import Enum
from typing import NamedTuple


class RelClass(NamedTuple):
    """Postgres objects."""

    rel_name: str
    is_readobject: bool
    is_readable: bool


class PGObject(RelClass, Enum):
    """RelClass object from relkind value."""

    r = RelClass("Relation table", True, True)
    i = RelClass("Index", False, False)
    S = RelClass("Sequence", False, False)
    t = RelClass("Toast table", False, False)
    v = RelClass("View", False, True)
    m = RelClass("Materialized view", False, True)
    c = RelClass("Composite type", False, False)
    f = RelClass("Foreign table", False, True)
    p = RelClass("Partitioned table", True, True)
    I = RelClass("Partitioned index", False, True)  # noqa: E741
    u = RelClass("Temporary table", True, True)
    o = RelClass("Optimized files", False, False)
    b = RelClass("Block directory", False, False)
    M = RelClass("Visibility map", False, False)

    def __str__(self) -> str:
        """String representation class."""

        return self.rel_name



from typing import (
    Iterable,
    NoReturn
)

from pgcopylib import PGCopyReader
from pgpack import (
    PGPackReader,
    metadata_reader,
)
from psycopg import Copy

from .reader import CopyReader


class StreamReader(PGPackReader):
    """Class for stream read from PostgreSQL/GreenPlum."""

    def __init__(
        self,
        metadata: bytes,
        copyobj: Iterable[Copy],
    ) -> None:
        """Class initialization."""

        self.metadata = metadata
        self.copyobj = CopyReader(copyobj)
        (
            self.columns,
            self.pgtypes,
            self.pgparam,
        ) = metadata_reader(self.metadata)
        self.pgcopy = PGCopyReader(
            self.copyobj,
            self.pgtypes,
        )

    def __str__(self) -> str:
        """String representation of PGPackReader."""

        def to_col(text: str) -> str:
            """Format string element."""

            text = text[:14] + "…" if len(text) > 15 else text
            return f" {text: <15} "

        empty_line = (
            "│-----------------+-----------------│"
        )
        end_line = (
            "└─────────────────┴─────────────────┘"
        )
        _str = [
            "<PostgreSQL/GreenPlum stream reader>",
            "┌─────────────────┬─────────────────┐",
            "│ Column Name     │ PostgreSQL Type │",
            "╞═════════════════╪═════════════════╡",
        ]

        for column, pgtype in zip(self.columns, self.pgtypes):
            _str.append(
                f"│{to_col(column)}│{to_col(pgtype.name)}│",
            )
            _str.append(empty_line)

        _str[-1] = end_line
        return "\n".join(_str) + f"""
Total columns: {len(self.columns)}
Readed rows: {self.pgcopy.num_rows}
"""

    def to_bytes(self) -> NoReturn:
        """Get raw unpacked pgcopy data."""

        raise NotImplementedError("Don't support in stream mode.")

    def close(self) -> None:
        """Close stream object."""

        self.copyobj.close()



from pathlib import Path
from random import randbytes
from re import match


pattern = r"\(select \* from (.*)\)|(.*)"


def search_object(table: str, query: str = "") -> str:
    """Return current string for object."""

    if query:
        return "query"

    return match(pattern, table).group(1) or table


def random_name() -> str:
    """Generate random name for prepare and temp table."""

    return f"session_{randbytes(8).hex()}"  # noqa: S311


def query_path() -> str:
    """Path for queryes."""

    return f"{Path(__file__).parent.absolute()}/queryes/{{}}.sql"


def query_template(query_name: str) -> str:
    """Get query template for his name."""

    path = query_path().format(query_name)

    with open(path, encoding="utf-8") as query:
        return query.read()


def chunk_query(query: str | None) -> tuple[list[str]]:
    """Chunk multiquery to queryes."""

    if not query:
        return [], []

    first_part: list[str] = [
        part.strip()
        for part in query.split(";")
    ]
    second_part: list[str] = []

    for _ in first_part:
        second_part.append(first_part.pop())
        if any(
            word == second_part[-1][:len(word)].lower()
            for word in ("with", "select")
        ):
            second_part = list(reversed(second_part))
            break

    return first_part, second_part





from io import (
    BufferedReader,
    BufferedWriter,
)
from logging import Logger
from types import MethodType
from typing import (
    Any,
    Iterable,
    Union
)

from pgcopylib import PGCopyWriter
from pgpack import (
    CompressionMethod,
    PGPackReader,
    PGPackWriter,
    metadata_reader,
)
from psycopg import (
    Connection,
    Cursor,
)
from pandas import DataFrame as PdFrame
from polars import DataFrame as PlFrame
from sqlparse import format as sql_format

from .common import (
    CopyBuffer,
    DumperLogger,
    PGConnector,
    PGPackDumperError,
    PGPackDumperReadError,
    PGPackDumperWriteError,
    PGPackDumperWriteBetweenError,
    StreamReader,
    chunk_query,
)


class PGPackDumper:
    """Class for read and write PGPack format."""

    def __init__(
        self,
        connector: PGConnector,
        compression_method: CompressionMethod = CompressionMethod.ZSTD,
        logger: Logger | None = None,
    ) -> None:
        """Class initialization."""

        if not logger:
            logger = DumperLogger()

        try:
            self.connector: PGConnector = connector
            self.compression_method: CompressionMethod = compression_method
            self.logger = logger
            self.connect: Connection = Connection.connect(
                **self.connector._asdict()
            )
            self.cursor: Cursor = self.connect.cursor()
            self.copy_buffer: CopyBuffer = CopyBuffer(self.cursor, self.logger)
        except Exception as error:
            self.logger.error(f"{error.__class__.__name__}: {error}")
            raise PGPackDumperError(error)

        self.version = (
            f"{self.connect.info.server_version // 10000}."
            f"{self.connect.info.server_version % 1000}"
        )
        self.logger.info(
            f"PGPackDumper initialized for host {self.connector.host}"
            f"[version {self.version}]"
        )

    @staticmethod
    def multiquery(dump_method: MethodType):
        """Multiquery decorator."""

        def wrapper(*args, **kwargs):

            first_part: list[str]
            second_part: list[str]

            self: PGPackDumper = args[0]
            cursor: Cursor = kwargs.get("dumper_src", self).cursor
            query: str = kwargs.get("query_src") or kwargs.get("query")
            part: int = 1
            first_part, second_part = chunk_query(self.query_formatter(query))
            total_prts = len(sum((first_part, second_part), [])) or 1

            if first_part:
                self.logger.info("Multiquery detected.")

                for query in first_part:
                    self.logger.info(f"Execute query {part}/{total_prts}")
                    cursor.execute(query)
                    part += 1

            if second_part:
                for key in ("query", "query_src"):
                    if key in kwargs:
                        kwargs[key] = second_part.pop(0)
                        break

            self.logger.info(
                f"Execute stream {part}/{total_prts} [pgcopy mode]"
            )
            output = dump_method(*args, **kwargs)

            if second_part:
                for query in second_part:
                    part += 1
                    self.logger.info(f"Execute query {part}/{total_prts}")
                    cursor.execute(query)

            if output:
                self.refresh()

            return output

        return wrapper

    def query_formatter(self, query: str) -> str | None:
        """Reformat query."""

        if not query:
            return
        return sql_format(sql=query, strip_comments=True).strip().strip(";")

    @multiquery
    def read_dump(
        self,
        fileobj: BufferedWriter,
        query: str | None = None,
        table_name: str | None = None,
    ) -> None:
        """Read PGPack dump from PostgreSQL/GreenPlum."""

        try:
            self.copy_buffer.query = query
            self.copy_buffer.table_name = table_name
            pgpack = PGPackWriter(
                fileobj,
                self.copy_buffer.metadata,
                self.compression_method,
            )

            with self.copy_buffer.copy_to() as copy_to:
                pgpack.from_bytes(bytes(data) for data in copy_to)

            size = pgpack.tell()
            pgpack.close()
            self.logger.info(f"Successfully read {size} bytes.")
            self.logger.info(
                f"Read pgpack dump from {self.connector.host} done."
            )
            return True
        except Exception as error:
            self.logger.error(f"{error.__class__.__name__}: {error}")
            raise PGPackDumperReadError(error)

    def write_dump(
        self,
        fileobj: BufferedReader,
        table_name: str,
    ) -> None:
        """Write PGPack dump into PostgreSQL/GreenPlum."""

        try:
            pgpack = PGPackReader(fileobj)
            self.copy_buffer.table_name = table_name
            self.copy_buffer.copy_from(pgpack.to_bytes())
            self.connect.commit()
            size = pgpack.tell()
            self.logger.info(f"Successfully sending {size} bytes.")
            pgpack.close()
            self.refresh()
        except Exception as error:
            self.logger.error(f"{error.__class__.__name__}: {error}")
            raise PGPackDumperWriteError(error)

    @multiquery
    def write_between(
        self,
        table_dest: str,
        table_src: str | None = None,
        query_src: str | None = None,
        dumper_src: Union["PGPackDumper", object] = None,
    ) -> None:
        """Write from PostgreSQL/GreenPlum into PostgreSQL/GreenPlum."""

        try:
            if not dumper_src:
                connect = Connection.connect(**self.connector._asdict())
                self.logger.info(
                    f"Set new connection for host {self.connector.host}."
                )
                source_copy_buffer = CopyBuffer(
                    connect.cursor(),
                    self.logger,
                    query_src,
                    table_src,
                )
            elif dumper_src.__class__ is PGPackDumper:
                source_copy_buffer = dumper_src.copy_buffer
                source_copy_buffer.table_name = table_src
                source_copy_buffer.query = query_src
            else:
                reader = dumper_src.to_reader(
                    query=query_src,
                    table_name=table_src,
                )
                dtype_data = reader.to_rows()
                self.from_rows(
                    dtype_data=dtype_data,
                    table_name=table_dest,
                )
                size = reader.tell()
                self.logger.info(f"Successfully sending {size} bytes.")
                return reader.close()

            self.copy_buffer.table_name = table_dest
            self.copy_buffer.copy_between(source_copy_buffer)
            self.connect.commit()
            return True
        except Exception as error:
            self.logger.error(f"{error.__class__.__name__}: {error}")
            raise PGPackDumperWriteBetweenError(error)

    @multiquery
    def to_reader(
        self,
        query: str | None = None,
        table_name: str | None = None,
    ) -> StreamReader:
        """Get stream from PostgreSQL/GreenPlum as StreamReader object."""

        self.copy_buffer.query = query
        self.copy_buffer.table_name = table_name
        return StreamReader(
            self.copy_buffer.metadata,
            self.copy_buffer.copy_to(),
        )

    def from_rows(
        self,
        dtype_data: Iterable[Any],
        table_name: str,
    ) -> None:
        """Write from python iterable object
        into PostgreSQL/GreenPlum table."""

        self.copy_buffer.table_name = table_name
        _, pgtypes, _ = metadata_reader(self.copy_buffer.metadata)
        writer = PGCopyWriter(None, pgtypes)
        self.copy_buffer.copy_from(writer.from_rows(dtype_data))
        self.connect.commit()
        self.refresh()

    def from_pandas(
        self,
        data_frame: PdFrame,
        table_name: str,
    ) -> None:
        """Write from pandas.DataFrame into PostgreSQL/GreenPlum table."""

        self.from_rows(
            dtype_data=iter(data_frame.values),
            table_name=table_name,
        )

    def from_polars(
        self,
        data_frame: PlFrame,
        table_name: str,
    ) -> None:
        """Write from polars.DataFrame into PostgreSQL/GreenPlum table."""

        self.from_rows(
            dtype_data=data_frame.iter_rows(),
            table_name=table_name,
        )

    def refresh(self) -> None:
        """Refresh session."""

        self.connect = Connection.connect(**self.connector._asdict())
        self.cursor = self.connect.cursor()
        self.copy_buffer.cursor = self.cursor
        self.logger.info(f"Connection to host {self.connector.host} updated.")

    def close(self) -> None:
        """Close session."""

        self.cursor.close()
        self.connect.close()
        self.logger.info(f"Connection to host {self.connector.host} closed.")



from datetime import datetime
from logging import (
    DEBUG,
    FileHandler,
    Formatter,
    Logger,
    StreamHandler,
)
from os import makedirs
from os.path import dirname
from sys import stdout

from ..version import __version__


def root_dir() -> str:
    """Get project directory."""

    try:
        import __main__

        return dirname(__main__.__file__)
    except AttributeError:
        return ""


class DumperLogger(Logger):
    """PGPackDumper logger."""

    def __init__(
            self,
            level: int = DEBUG,
            use_console: bool = True,
    ) -> None:
        """Class initialize."""

        super().__init__("PGPackDumper")

        self.fmt = (
            f"%(asctime)s | %(levelname)-8s | ver {__version__} "
            "| %(funcName)s-%(filename)s-%(lineno)04d <%(message)s>"
        )
        self.setLevel(level)
        self.log_path = f"{root_dir()}/pgpack_logs"
        makedirs(self.log_path, exist_ok=True)

        formatter = Formatter(
            fmt=self.fmt,
            datefmt="%Y-%m-%d %H:%M:%S",
        )

        file_handler = FileHandler(
            "{}/{:%Y-%m-%d}_{}.log".format(
                self.log_path,
                datetime.now(),
                self.name,
            ),
            encoding="utf-8",
        )
        file_handler.setLevel(DEBUG)
        file_handler.setFormatter(formatter)
        self.addHandler(file_handler)

        if use_console:
            console_handler = StreamHandler(stdout)
            console_handler.setLevel(level)
            console_handler.setFormatter(formatter)
            self.addHandler(console_handler)

        self.propagate = False



from setuptools import (
    Extension,
    setup,
)
from Cython.Build import cythonize


extensions = [
    Extension(
        "pgpack_dumper.common.reader",
        ["src/pgpack_dumper/common/reader.pyx"],
    ),
]


setup(
    name="pgpack_dumper",
    package_dir={"": "src"},
    ext_modules=cythonize(extensions, language_level="3"),
    packages=[
        "pgpack_dumper.common",
    ],
    package_data={
        "pgpack_dumper": [
            "**/*.pyx",
            "**/*.pxd",
            "**/*.sql",
            "*.pxd",
            "*.pyd",
            "*.md",
            "*.txt",
            "*.sql",
        ]
    },
    exclude_package_data={
        "": ["**/*.c", "*.c"],
        "pgpack_dumper": ["**/*.c", "*.c"],
    },
    include_package_data=True,
    setup_requires=["Cython>=3.0"],
    zip_safe=False,
)



from logging import Logger
from typing import (
    Generator,
    Iterator,
)

from psycopg import (
    Copy,
    Cursor,
)

from .errors import (
    CopyBufferObjectError,
    CopyBufferTableNotDefined,
)
from .query import (
    query_template,
    search_object,
)
from .structs import PGObject
from .metadata import read_metadata


class CopyBuffer:

    def __init__(
        self,
        cursor: Cursor,
        logger: Logger,
        query: str | None = None,
        table_name: str | None = None,
    ) -> None:
        """Class initialization."""

        self.cursor = cursor
        self.logger = logger
        self.query = query
        self.table_name = table_name

    @property
    def metadata(self) -> bytes:
        """Get metadata as bytes."""

        host = self.cursor.connection.info.host
        self.logger.info(f"Start read metadata from host {host}.")
        metadata = read_metadata(
            self.cursor,
            self.query,
            self.table_name,
        )
        self.logger.info(f"Read metadata from host {host} done.")
        return metadata

    def copy_to(self) -> Iterator[Copy]:
        """Get copy object from PostgreSQL."""

        if not self.query and not self.table_name:
            error_msg = "Query or table not defined."
            self.logger.error(f"CopyBufferTableNotDefined: {error_msg}")
            raise CopyBufferTableNotDefined(error_msg)

        host = self.cursor.connection.info.host

        if not self.query:
            self.logger.info(f"Start read from {host}.{self.table_name}.")
            self.cursor.execute(query_template("relkind").format(
                table_name=self.table_name,
            ))
            relkind = self.cursor.fetchone()[0]
            pg_object = PGObject[relkind]
            if not pg_object.is_readable:
                error_msg = f"Read from {pg_object} not support."
                self.logger.error(f"CopyBufferObjectError: {error_msg}")
                raise CopyBufferObjectError(error_msg)
            self.logger.info(f"Use method read from {pg_object}.")
            if not pg_object.is_readobject:
                self.table_name = f"(select * from {self.table_name})"
        elif self.query:
            self.logger.info(f"Start read query from {host}.")
            self.logger.info("Use method read from select.")
            self.table_name = f"({self.query})"

        return self.cursor.copy(
            query_template("copy_to").format(table_name=self.table_name)
        )

    def copy_from(
        self,
        copyobj: Iterator[bytes],
    ) -> None:
        """Write PGCopy dump into PostgreSQL."""

        if not self.table_name:
            error_msg = "Table not defined."
            self.logger.error(f"CopyBufferTableNotDefined: {error_msg}")
            raise CopyBufferTableNotDefined(error_msg)

        host = self.cursor.connection.info.host
        size = 0
        self.logger.info(f"Start write into {host}.{self.table_name}.")

        with self.cursor.copy(
            query_template("copy_from").format(table_name=self.table_name)
        ) as cp:
            for bytes_data in copyobj:
                size += len(bytes_data)
                cp.write(bytes_data)

        self.logger.info(f"Successfully sending {size} bytes.")
        self.logger.info(f"Write into {host}.{self.table_name} done.")

    def copy_between(
        self,
        copy_buffer: "CopyBuffer",
    ) -> None:
        """Write from PostgreSQL into PostgreSQL."""

        with copy_buffer.copy_to() as copy_to:
            destination_host = self.cursor.connection.info.host
            source_host = copy_buffer.cursor.connection.info.host
            source_object = search_object(
                copy_buffer.table_name,
                copy_buffer.query,
            )
            size = 0
            self.logger.info(
                f"Copy {source_object} from {source_host} into "
                f"{destination_host}.{self.table_name} started."
            )

            with self.cursor.copy(
                query_template("copy_from").format(table_name=self.table_name)
            ) as copy_from:
                for data in copy_to:
                    size += len(data)
                    copy_from.write(data)

            self.logger.info(f"Successfully sending {size} bytes.")
            self.logger.info(
                f"Copy {source_object} from {source_host}"
                f"into {destination_host}.{self.table_name} done."
            )

    def copy_reader(self) -> Generator[bytes, None, None]:
        """Read bytes from copy object."""

        host = self.cursor.connection.info.host
        source = search_object(
            self.table_name,
            self.query,
        )

        with self.copy_to() as copy_object:
            for data in copy_object:
                yield bytes(data)

        self.logger.info(f"Read {source} from {host} done.")



class CopyBufferError(Exception):
    """CopyBuffer base error."""


class CopyBufferObjectError(TypeError):
    """Destination object not support."""


class CopyBufferTableNotDefined(ValueError):
    """Destination table not defined."""


class PGPackDumperError(Exception):
    """PGPackDumper base error."""


class PGPackDumperReadError(PGPackDumperError):
    """PGPackDumper read error."""


class PGPackDumperWriteError(PGPackDumperError):
    """PGPackDumper write error."""


class PGPackDumperWriteBetweenError(PGPackDumperWriteError):
    """PGPackDumper write between error."""
